@@ -6,36 +6,54 @@ import re
 import warnings
 import weakref
 from abc import ABC, abstractmethod
+from functools import cached_property
+from pathlib import Path
+from typing import TYPE_CHECKING, Final, Literal
 
 import inifix
 import numpy as np
 
-from yt.data_objects.index_subobjects.grid_patch import AMRGridPatch
+from yt.data_objects.index_subobjects.stretched_grid import StretchedGrid
 from yt.data_objects.static_output import Dataset
 from yt.funcs import setdefaultattr
 from yt.geometry.grid_geometry_handler import GridIndex
+from yt.utilities.lib.misc_utilities import _obtain_coords_and_widths
+from yt_idefix._typing import UnitLike
 
 from ._io import C_io, dmp_io, vtk_io
 from ._io.commons import IdefixFieldProperties, IdefixMetadata
 from .definitions import _PlutoBaseUnits, pluto_def_constants
-from .fields import IdefixDmpFieldInfo, IdefixVtkFieldInfo
+from .fields import BaseVtkFields, IdefixDmpFields, IdefixVtkFields, PlutoVtkFields
+
+# import IO classes to ensure they are properly registered,
+# even though we don't call them directly
+from .io import IdefixDmpIO, IdefixVtkIO, PlutoVtkIO  # noqa
+
+if TYPE_CHECKING:
+    # these should really be unyt_array,
+    # but mypy doesn't recognize it as a valid type as of unyt 2.9.3 and mypy 0.991
+    XSpans = np.ndarray
+    YSpans = np.ndarray
+    ZSpans = np.ndarray
+    XCoords = np.ndarray
+    YCoords = np.ndarray
+    ZCoords = np.ndarray
 
 ytLogger = logging.getLogger("yt")
+_DEF_GEOMETRY_REGEXP: Final = re.compile(r"^\s*#define\s+GEOMETRY\s+([A-Z]+)")
+_DEF_UNIT_REGEXP: Final = re.compile(r"^\s*#define\s+UNIT_(\w+)\s+(\S+)")
 
 
-class IdefixGrid(AMRGridPatch):
+class IdefixGrid(StretchedGrid):
     _id_offset = 0
 
-    def __init__(self, id, index, level, dims):
-        super().__init__(id, filename=index.index_filename, index=index)
+    def __init__(self, id, cell_widths, filename, index, level, dims):
+        super().__init__(id=id, filename=filename, index=index, cell_widths=cell_widths)
         self.Parent = None
         self.Children = []
         self.Level = level
         self.ActiveDimensions = dims
 
-    def __repr__(self):
-        return "IdefixGrid_%04i (%s)" % (self.id, self.ActiveDimensions)
-
 
 class IdefixHierarchy(GridIndex, ABC):
     grid = IdefixGrid
@@ -63,8 +81,10 @@ class IdefixHierarchy(GridIndex, ABC):
         self.grid_right_edge[0][:] = self.ds.domain_right_edge[:]
         self.grid_dimensions[0][:] = self.ds.domain_dimensions[:]
         self.grid_particle_count[0][0] = 0
-        self.grid_levels[0][0] = 1
-        self.max_level = 1
+
+        # Idefix/Pluto are not AMR
+        self.grid_levels[0][0] = 0
+        self.min_level = self.max_level = 0
 
         self._field_offsets = self._get_field_offset_index()
 
@@ -80,11 +100,21 @@ class IdefixHierarchy(GridIndex, ABC):
         #   g.Parent   <= parent grid
         # This is handled by the frontend because often the children must be identified.
         self.grids = np.empty(self.num_grids, dtype="object")
-        for i in range(self.num_grids):
-            g = self.grid(i, self, self.grid_levels.flat[i], self.grid_dimensions[i])
-            g._prepare_grid()
-            g._setup_dx()
-            self.grids[i] = g
+
+        assert self.num_grids == 1
+
+        i = 0
+        g = self.grid(
+            id=i,
+            index=self,
+            filename=self.index_filename,
+            cell_widths=self._cell_widths,
+            level=self.grid_levels.flat[i],
+            dims=self.grid_dimensions[i],
+        )
+        g._prepare_grid()
+        g._setup_dx()
+        self.grids[i] = g
 
     @abstractmethod
     def _get_field_offset_index(self) -> dict[str, int]:
@@ -94,53 +124,182 @@ class IdefixHierarchy(GridIndex, ABC):
             field_index = ...
         return field_index  # type: ignore
 
+    @abstractmethod
+    @cached_property
+    def _cell_widths(self) -> tuple[XSpans, YSpans, ZSpans]:
+        # must return a 3-tuple of 1D unyt_array
+        # with unit "code_length" and dtype float64
+        ...
+
+    @abstractmethod
+    @cached_property
+    def _cell_centers(self) -> tuple[XCoords, YCoords, ZCoords]:
+        # must return a 3-tuple of 1D unyt_array
+        # with unit "code_length" and dtype float64
+        ...
+
+    def _icoords_to_fcoords(
+        self,
+        icoords: np.ndarray,
+        ires: np.ndarray,
+        axes: tuple[int, ...] | None = None,
+    ):
+        if axes is None:
+            axes = (0, 1, 2)
+        # this is needed to support projections
+        coords = np.empty(icoords.shape, dtype="f8")
+        cell_widths = np.empty(icoords.shape, dtype="f8")
+        for i, ax in enumerate(axes):
+            coords[:, i], cell_widths[:, i] = _obtain_coords_and_widths(
+                icoords[:, i],
+                ires,
+                self._cell_widths[ax],
+                self.ds.domain_left_edge[ax].d,
+            )
+        return coords, cell_widths
+
 
 class IdefixVtkHierarchy(IdefixHierarchy):
     def _get_field_offset_index(self) -> dict[str, int]:
         return self.ds._field_offset_index
 
+    @cached_property
+    def _cell_widths(self) -> tuple[XSpans, YSpans, ZSpans]:
+        with open(self.index_filename, "rb") as fh:
+            cell_edges = vtk_io.read_grid_coordinates(fh, geometry=self.ds.geometry)
+
+        dims = self.ds.domain_dimensions
+        length_unit = self.ds.quan(1, "code_length")
+
+        cell_widths: tuple[XSpans, YSpans, ZSpans]
+        cell_widths = (
+            np.empty(max(dims[0], 2), dtype="float64") * length_unit,
+            np.empty(max(dims[1], 2), dtype="float64") * length_unit,
+            np.empty(max(dims[2], 2), dtype="float64") * length_unit,
+        )
+
+        for idir, edges in enumerate(cell_edges[:3]):
+            if dims[idir] > 1:
+                cell_widths[idir][:] = np.ediff1d(edges)
+            else:
+                cell_widths[idir][:] = self.ds.domain_width[idir]
+        return cell_widths
+
+    @cached_property
+    def _cell_centers(self) -> tuple[XCoords, YCoords, ZCoords]:
+        with open(self.index_filename, "rb") as fh:
+            cell_edges = vtk_io.read_grid_coordinates(fh, geometry=self.ds.geometry)
+
+        dims = self.ds.domain_dimensions
+        length_unit = self.ds.quan(1, "code_length")
+
+        cell_centers: tuple[XCoords, YCoords, ZCoords]
+        cell_centers = (
+            np.empty(max(dims[0], 2), dtype="float64") * length_unit,
+            np.empty(max(dims[1], 2), dtype="float64") * length_unit,
+            np.empty(max(dims[2], 2), dtype="float64") * length_unit,
+        )
+
+        for idir, edges in enumerate(cell_edges[:3]):
+            if dims[idir] > 1:
+                cell_centers[idir][:] = 0.5 * (edges[1:] + edges[:-1])
+            else:
+                cell_centers[idir][:] = edges[0]
+        return cell_centers
+
 
 class IdefixDmpHierarchy(IdefixHierarchy):
     def _get_field_offset_index(self) -> dict[str, int]:
         with open(self.index_filename, "rb") as fh:
             return dmp_io.get_field_offset_index(fh)
 
+    @cached_property
+    def _cell_widths(self) -> tuple[XSpans, YSpans, ZSpans]:
+        _fprops, fdata = dmp_io.read_idefix_dmpfile(self.index_filename, skip_data=True)
+        length_unit = self.ds.quan(1, "code_length")
+        return (
+            (fdata["xr1"] - fdata["xl1"]) * length_unit,
+            (fdata["xr2"] - fdata["xl2"]) * length_unit,
+            (fdata["xr3"] - fdata["xl3"]) * length_unit,
+        )
+
+    @cached_property
+    def _cell_centers(self) -> tuple[XCoords, YCoords, ZCoords]:
+        _fprops, fdata = dmp_io.read_idefix_dmpfile(self.index_filename, skip_data=True)
+        length_unit = self.ds.quan(1, "code_length")
+        return (
+            fdata["x1"] * length_unit,
+            fdata["x2"] * length_unit,
+            fdata["x3"] * length_unit,
+        )
+
 
 class IdefixDataset(Dataset, ABC):
     """A common abstraction for IdefixDmpDataset and IdefixVtkDataset."""
 
     _version_regexp = re.compile(r"v\d+\.\d+\.?\d*[-\w+]*")
+    _dataset_type: str  # defined in subclasses
+    _default_definitions_header = "definitions.hpp"
+    _default_inifile = "idefix.ini"
 
     def __init__(
         self,
         filename,
-        dataset_type=None,
-        unit_system="cgs",
-        units_override=None,
         *,
-        geometry: str | None = None,
-        inifile=None,
+        dataset_type: str | None = None,  # deleguated to child classes
+        units_override: dict[str, UnitLike] | None = None,
+        unit_system: Literal["cgs", "mks", "code"] = "cgs",
+        default_species_fields: Literal["neutral", "ionized"] | None = None,
+        # from here, frontend-specific arguments
+        geometry: Literal["cartesian", "spherical", "cylindrical", "polar"]
+        | None = None,
+        inifile: str | os.PathLike[str] | None = None,
+        definitions_header: str | os.PathLike[str] | None = None,
     ):
-        self._geometry_from_user: str | None = geometry
+        self._geometry_from_user = geometry
 
         dt = type(self)._dataset_type
         self.fluid_types += (dt,)
 
+        self._input_filename: str = os.fspath(filename)
+        self._inifile = self._get_meta_file(inifile, default=self._default_inifile)
+        self._definitions_header = self._get_meta_file(
+            definitions_header, default=self._default_definitions_header
+        )
+
         super().__init__(
             filename,
             dataset_type=dt,
             units_override=units_override,
             unit_system=unit_system,
+            default_species_fields=default_species_fields,
         )
 
-        self.inifile = inifile
-        self._parse_inifile()
-
         self.storage_filename = None
 
         # idefix does not support grid refinement
         self.refine_by = 1
 
+    def _get_meta_file(
+        self, arg: str | os.PathLike[str] | None, /, *, default: str
+    ) -> str:
+        root_dir = Path(self.directory)
+
+        if arg is not None:
+            if os.path.isabs(arg):
+                return os.fspath(arg)
+            else:
+                return str((root_dir / arg).absolute())
+
+        _, ext = os.path.splitext(default)
+        if (
+            len(candidates := list(root_dir.glob(f"*{ext}"))) == 1
+            and (file := candidates[0]).name == default
+        ):
+            return str(file.absolute())
+        else:
+            return ""
+
     def _parse_parameter_file(self):
         # base method, intended to be subclassed
         # parse the version hash
@@ -153,54 +312,64 @@ class IdefixDataset(Dataset, ABC):
         self.omega_matter = 0.0
         self.hubble_constant = 0.0
 
+        self._parse_inifile()
+        self._parse_definitions_header()
         self._setup_geometry()
 
     def _setup_geometry(self) -> None:
-        from_file = self.parameters.get("geometry")
-        from_user = self._geometry_from_user
-        if from_file is None and from_user is None:
+        from_bin = self.parameters.get("geometry", "")
+        from_definitions = self.parameters["definitions"].get("geometry", "")
+        from_input = self._geometry_from_user or ""
+
+        if from_definitions and from_bin and from_bin != from_definitions:
+            raise RuntimeError(
+                "Geometries from disk file and definitions header do not match, got\n"
+                f" - {from_bin!r} (from {self.parameter_filename})\n"
+                f" - {from_definitions!r} (from {self._definitions_header})"
+            )
+
+        from_disk = from_bin or from_definitions
+
+        if not any((from_disk, from_input)):
             raise ValueError(
-                "Geometry couldn't be parsed from file. "
+                "Geometry couldn't be parsed from disk. "
                 "The 'geometry' keyword argument must be specified."
             )
-        elif from_user is not None:
-            if from_file is not None and from_user != from_user:
+
+        if from_input:
+            if from_disk and from_input and from_input != from_disk:
                 warnings.warn(
-                    f"Got inconsistent geometry flags:\n"
-                    f" - {from_file!r} (from file)\n"
-                    f" - {from_user!r} (from user)\n"
-                    "user-input prevails to allow working around hypothetical parsing bugs, "
+                    "Geometries from disk and input do not match, got\n"
+                    f" - {from_disk!r} (from disk)\n"
+                    f" - {from_input!r} (from input)\n"
+                    "input prevails to allow working around hypothetical parsing bugs, "
                     "but it is very likely to result in an error in the general case."
                 )
-            self.geometry = from_user
+            self.geometry = from_input
         else:
-            assert from_file is not None
-            self.geometry = from_file
+            assert from_disk
+            self.geometry = from_disk
 
     def _parse_inifile(self) -> None:
-        if self.inifile is None:
+        if not self._inifile:
             return
 
-        self.parameters.update(inifix.load(self.inifile))
-        grid_ini = self.parameters["Grid"]
+        with open(self._inifile, "rb") as fh:
+            self.parameters.update(inifix.load(fh))
 
-        msg_elems: list[str] = []
-        for ax, vals in grid_ini.items():
-            if vals[0] > 1:
-                # more than one block is only relevant for mixing grid spacings,
-                # but only "u" is supported
-                msg_elems.append(f"found multiple blocks in direction {ax}; got {vals}")
-            if any(_ != "u" for _ in vals[3::3]):
-                msg_elems.append(f"found non-uniform block(s) in direction {ax}")
-        if len(msg_elems) > 0:
-            msg = (
-                "Streched grid detected !\n"
-                + "- "
-                + "\n- ".join(msg_elems)
-                + "\nThe grid will be treated as uniformly spaced in every direction. "
-                "yt_idefix.load_stretched may be a better alternative loader for these data"
-            )
-            warnings.warn(msg)
+    def _parse_definitions_header(self) -> None:
+        self.parameters["definitions"] = {}
+        if not self._definitions_header:
+            return
+
+        with open(self._definitions_header) as fh:
+            body = fh.read()
+        lines = C_io.strip_comments(body).split("\n")
+
+        for line in lines:
+            if (geom_match := re.fullmatch(_DEF_GEOMETRY_REGEXP, line)) is not None:
+                self.parameters["definitions"]["geometry"] = geom_match.group(1).lower()
+                return
 
     def _set_code_unit_attributes(self):
         # This is where quantities are created that represent the various
@@ -222,7 +391,7 @@ class IdefixDataset(Dataset, ABC):
     # The following methods are frontend-specific
 
     @abstractmethod
-    def _get_header(self) -> str:
+    def _read_data_header(self) -> str:
         pass
 
     def _get_code_version(self) -> str:
@@ -230,12 +399,11 @@ class IdefixDataset(Dataset, ABC):
         # - in Idefix dumps there's only one line
         # - in Vtk files (Idefix or Pluto), there are two,
         #   the first of which isn't code specific
-        header = self._get_header().splitlines()[-1]
+        header = self._read_data_header().splitlines()[-1]
 
         regexp = self.__class__._version_regexp
 
         match = re.search(regexp, header)
-        version: str
         if match is None:
             warnings.warn(
                 f"Could not determine code version from file header {header!r}"
@@ -247,7 +415,7 @@ class IdefixDataset(Dataset, ABC):
 
 class IdefixVtkDataset(IdefixDataset):
     _index_class = IdefixVtkHierarchy
-    _field_info_class = IdefixVtkFieldInfo
+    _field_info_class: type[BaseVtkFields] = IdefixVtkFields
     _dataset_type = "idefix-vtk"
     _required_header_keyword = "Idefix"
 
@@ -260,7 +428,7 @@ class IdefixVtkDataset(IdefixDataset):
         else:
             return cls._required_header_keyword in header
 
-    def _get_header(self) -> str:
+    def _read_data_header(self) -> str:
         return vtk_io.read_header(self.parameter_filename)
 
     def _parse_parameter_file(self):
@@ -293,7 +461,7 @@ class IdefixVtkDataset(IdefixDataset):
         self.domain_right_edge = dre
 
         # time wasn't stored in vtk files before Idefix 0.8
-        self.current_time = self.quan(md.get("time", -1), "code_time")
+        self.current_time = md.get("time", -1)
 
         # periodicity was not stored in vtk files before Idefix 0.9
         self._periodicity = md.get("periodicity", (True, True, True))
@@ -301,25 +469,22 @@ class IdefixVtkDataset(IdefixDataset):
 
 class IdefixDmpDataset(IdefixDataset):
     _index_class = IdefixDmpHierarchy
-    _field_info_class = IdefixDmpFieldInfo
+    _field_info_class = IdefixDmpFields
     _dataset_type = "idefix-dmp"
 
     @classmethod
-    def _is_valid(cls, filename, *args, **kwargs):
-        ok = bool(
-            re.match(r"^(dump)\.\d{4}(\.dmp)$", os.path.basename(filename))
-        )  # this is possibly too restrictive
+    def _is_valid(cls, filename, *args, **kwargs) -> bool:
         try:
-            ok &= "Idefix" in dmp_io.read_header(filename)
+            header_string = dmp_io.read_header(filename)
+            return re.match(r"Idefix .* Dump Data", header_string) is not None
         except Exception:
-            ok = False
-        return ok
+            return False
 
     def _get_fields_metadata(self) -> tuple[IdefixFieldProperties, IdefixMetadata]:
         # read everything except large arrays
         return dmp_io.read_idefix_dmpfile(self.parameter_filename, skip_data=True)
 
-    def _get_header(self) -> str:
+    def _read_data_header(self) -> str:
         return dmp_io.read_header(self.parameter_filename)
 
     def _parse_parameter_file(self):
@@ -352,108 +517,87 @@ class IdefixDmpDataset(IdefixDataset):
 
 
 class PlutoVtkDataset(IdefixVtkDataset):
+    _field_info_class = PlutoVtkFields
+    _dataset_type = "pluto-vtk"
     _version_regexp = re.compile(r"\d+\.\d+\.?\d*[-\w+]*")
     _required_header_keyword = "PLUTO"
-
-    def __init__(
-        self,
-        filename,
-        dataset_type=None,
-        unit_system="cgs",
-        units_override=None,
-        *,
-        geometry: str | None = None,
-        definitions_header: str | None = None,
-        inifile=None,
-    ):
-        self._definitions_header: str | None
-        if definitions_header is not None:
-            self._definitions_header = os.fspath(definitions_header)
-        else:
-            self._definitions_header = None
-
-        super().__init__(
-            filename,
-            dataset_type=dataset_type,
-            unit_system=unit_system,
-            units_override=units_override,
-            geometry=geometry,
-            inifile=inifile,
-        )
+    _default_definitions_header = "definitions.h"
+    _default_inifile = "pluto.ini"
 
     def _parse_parameter_file(self):
-        self._parse_header_file()
         super()._parse_parameter_file()
-        self._get_time()
 
-    def _parse_header_file(self):
-        """Read some metadata from header file 'definitions.h'."""
-        geom_regexp = re.compile(r"^\s*#define\s+GEOMETRY\s+([A-Z]+)")
-        unit_regexp = re.compile(r"^\s*#define\s+UNIT_(\w+)\s+(\S+)")
-        constexpr = re.compile(r"CONST_\w+")
-
-        # definitions.h is presumed to be along with data file
-        if self._definitions_header is None:
-            self._definitions_header = os.path.join(self.fullpath, "definitions.h")
-        elif not os.path.isfile(self._definitions_header):
-            raise FileNotFoundError(f"No such file {self._definitions_header!r}")
-
-        if os.path.isfile(self._definitions_header):
-            with open(self._definitions_header) as fh:
-                body = fh.read()
-            lines = C_io.strip_comments(body).split("\n")
-
-            for line in lines:
-                geom_match = re.fullmatch(geom_regexp, line)
-                if geom_match is not None:
-                    self.parameters["geometry"] = geom_match.group(1).lower()
-                    continue
-
-                unit_match = re.fullmatch(unit_regexp, line)
-                if unit_match is not None:
-                    unit = unit_match.group(1).lower() + "_unit"
-                    expr = unit_match.group(2)
-                    expr = re.sub(constexpr, self._get_constants, expr)
-                    self.parameters[unit] = eval(expr)
-        else:
-            warnings.warn(
-                f"Header file {self._definitions_header} couldn't be found. "
-                "The code units are set to be 1.0 in cgs by default."
-            )
-
-    def _get_constants(self, match: re.Match) -> str:
-        """Replace matched constant string with its value"""
-        key = match.group()
-        return str(pluto_def_constants[key])
-
-    def _get_time(self):
-        """Get current time from vtk.out."""
-        log_file = os.path.join(self.fullpath, "vtk.out")
-        match = re.search(r"\.(\d*)\.", self.parameter_filename)
-        if match is None:
+        # parse time from vtk.out
+        log_file = os.path.join(self.directory, "vtk.out")
+        if (match := re.search(r"\.(\d*)\.", self.parameter_filename)) is None:
             raise RuntimeError(
                 f"Failed to parse output number from file name {self.parameter_filename}"
             )
         index = int(match.group(1))
 
-        self.current_time = self.quan(-1, "code_time")
-        if os.path.isfile(log_file):
-            log_regexp = re.compile(rf"^{index}\s(\S+)")
-            with open(log_file) as fh:
-                for line in fh.readlines():
-                    log_match = re.search(log_regexp, line)
-                    if log_match:
-                        self.current_time = self.quan(
-                            float(log_match.group(1)), "code_time"
-                        )
-                        break
-                else:
-                    ytLogger.warning(
-                        "Failed to retrieve time from %s, setting current_time = -1",
-                        log_file,
-                    )
-        else:
+        # will be converted to actual unyt_quantity in _set_derived_attrs
+        self.current_time = -1
+
+        if not os.path.isfile(log_file):
             ytLogger.warning("Missing log file %s, setting current_time = -1", log_file)
+            return
+
+        log_regexp = re.compile(rf"^{index}\s(\S+)")
+        with open(log_file) as fh:
+            for line in fh.readlines():
+                log_match = re.search(log_regexp, line)
+                if log_match:
+                    self.current_time = float(log_match.group(1))
+                    break
+            else:
+                ytLogger.warning(
+                    "Failed to retrieve time from %s, setting current_time = -1",
+                    log_file,
+                )
+
+    def _parse_definitions_header(self) -> None:
+        """Read some metadata from header file 'definitions.h'."""
+        self.parameters["definitions"] = {}
+        if not self._definitions_header:
+            ytLogger.warning(
+                "%s was not found. Code units will be set to 1.0 in cgs.",
+                self._default_definitions_header,
+            )
+            return
+
+        with open(self._definitions_header) as fh:
+            body = fh.read()
+        lines = C_io.strip_comments(body).split("\n")
+
+        for line in lines:
+            if (geom_match := re.fullmatch(_DEF_GEOMETRY_REGEXP, line)) is not None:
+                self.parameters["definitions"]["geometry"] = geom_match.group(1).lower()
+            elif (unit_match := re.fullmatch(_DEF_UNIT_REGEXP, line)) is not None:
+                unit = unit_match.group(1).lower() + "_unit"
+                expr = unit_match.group(2)
+                # Before evaluating the expression, replace the input parameters,
+                # pre-defined constants, code units and sqrt function
+                # that cannot be resolved. The order doesn't matter.
+                expr = re.sub(r"g_inputParam\[(\w+)\]", self._get_input_parameter, expr)
+                expr = re.sub(r"CONST_\w+", self._get_constants, expr)
+                expr = re.sub(r"UNIT_(\w+)", self._get_unit, expr)
+                expr = re.sub(r"sqrt", "np.sqrt", expr)
+                self.parameters["definitions"][unit] = eval(expr)
+
+    def _get_input_parameter(self, match: re.Match) -> str:
+        """Replace matched input parameters with its value"""
+        key = match.group(1)
+        return str(self.parameters["Parameters"][key])
+
+    def _get_unit(self, match: re.Match) -> str:
+        """Replace matched unit with its value"""
+        key = match.group(1).lower() + "_unit"
+        return str(self.parameters["definitions"].get(key, 1.0))
+
+    def _get_constants(self, match: re.Match) -> str:
+        """Replace matched constant string with its value"""
+        key = match.group()
+        return str(pluto_def_constants[key])
 
     def _set_code_unit_attributes(self):
         """Conversion between physical units and code units."""
@@ -465,14 +609,11 @@ class PlutoVtkDataset(IdefixVtkDataset):
         # Default values of Pluto's base units which are stored in self.parameters
         # if they can be read from definitions.h
         # Otherwise, they are set to unity in cgs.
+        defs = self.parameters["definitions"]
         pluto_units = {
-            "velocity_unit": self.quan(
-                self.parameters.get("velocity_unit", 1.0), "cm/s"
-            ),
-            "density_unit": self.quan(
-                self.parameters.get("density_unit", 1.0), "g/cm**3"
-            ),
-            "length_unit": self.quan(self.parameters.get("length_unit", 1.0), "cm"),
+            "velocity_unit": self.quan(defs.get("velocity_unit", 1.0), "cm/s"),
+            "density_unit": self.quan(defs.get("density_unit", 1.0), "g/cm**3"),
+            "length_unit": self.quan(defs.get("length_unit", 1.0), "cm"),
         }
 
         uo_size = len(self.units_override)
